// comment of if you do not use/link to HDF5
#define ARMA_USE_HDF5 
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <iostream>
#include <boost/bind.hpp>
#include <boost/program_options.hpp>
#include "../mlstub/load.hpp"
#include "cvmtrain.h"
#include "icvmtrain.h"
#include "kernel-matrix.h"
#include "multiclassmodel_factory.h" 
#include "cvmmulticlassmodel.h"
#include "cvmmulticlassmodel_each.h"
#include <assert.h>
#include <fstream>
#include <boost/serialization/serialization.hpp>
#include <boost/shared_ptr.hpp>
#include "nystroem-kernel-matrix.h"
//#include "cvm_serialization.hpp"

using namespace std;
using namespace arma;
namespace po   = boost::program_options;

template<class Matrix>
void print_matrix(Matrix matrix) {
//    matrix.print(std::cout);
	std::ofstream ofstr("output.txt");
	matrix.print(ofstr);
}

//provide explicit instantiations of the template function for 
//every matrix type you use somewhere in your program.
// in gdb use call print_matrix<arma::Mat<double> >(matrix) 
template void print_matrix<arma::mat>(arma::mat matrix);
template void print_matrix<arma::cx_mat>(arma::cx_mat matrix);

template<typename T>
T mod(T a, int n)
{
    return a - floor(a/n)*n;
}   

int main(int argc, char** argv)
{
    // command line processing
	std::string sDataFilename = "./data/iris_n_3c.csv";
	std::string sModelFilename;
	sModelFilename.clear();
	unsigned int kFold        = 5;
	bool bCrossvalidate       = false;
	bool bPredict	          = false;
	bool bNystroem	          = false;	
	bool bIndefinite          = false;
	bool bHDF5	          = false;
	bool bSparse		  = false;
	unsigned int iCycles  	  = 500;
	int iKernel               = 0;
	double dC                 = 1;
	double dEpsilon		  = 0.01;
	int iSparse               = 10;
	int iMultiClassMode       = 0;
	bool bResult = false;
    try {
        po::options_description desc("Allowed options");
        desc.add_options()
            ("help", "produce help message")                                            
            ("data", po::value(&sDataFilename),"data filename (csv, last column labels)")            
			("k-fold", po::value(&kFold), "k for k-fold 0 (not set)")	    	    
		    	("cv", po::value(&bCrossvalidate), "crossvalidation (default false)")	    
			("cycles", po::value(&iCycles), "cycles (default 500)")
			("model", po::value(&sModelFilename), "model filename")
			("kernel", po::value(&iKernel), "kernel (0-linear (default), 1-elm-rbf, 2-input is a kernel")
			("nystroem", po::value(&bNystroem), "nystroem kernel (false)")
			("indefinite", po::value(&bIndefinite), "indefinite kernel (false)")
			("sparse", po::value(&bSparse), "sparse indefinite kernel (false)")
			("iSparsity", po::value(&iSparse), "sparsity level (e.g. [1,N] e.g. 10 (default))")
			("predict", po::value(&bPredict), "prediction (default false)")
			("hdf5", po::value(&bHDF5), "use hdf5 (binary) as input file format")
			("C", po::value(&dC), "C=1 (default)")
			("eps", po::value(&dEpsilon), "dEpsilon=0.01 (default)")
			("multiclassmode", po::value(&iMultiClassMode), "use 0 (one vs rest, default), 1 (each against each)")
        ;

        po::variables_map vm;        
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);    

        if (vm.count("help")) {
            std::cout << desc << "\n";
            return 0;
        }    
	        
	std::cout << "Current settings ..." << std::endl;            		          				
	std::cout << "Data file: " << sDataFilename << std::endl;            					    
	std::cout << "Model file: " << sModelFilename << std::endl;
	std::cout << "Cycles:  " << iCycles << std::endl;
	if(bSparse)
		std::cout << "Sparsity:  " << iSparse << std::endl;

	if(kFold > 1)
	{
	    std::cout << "k k-fold cv:  " << kFold << std::endl;	
	}

	if(iKernel < 0 || iKernel > 2)
	{
	        std::cerr << "unsupported kernel!\n";
		return -1;
	}

	if(bNystroem && bCrossvalidate)
	{
		bCrossvalidate = false;
		std::cout << "Crossvalidation on Nystroem approximated matrix not yet supported" << std::endl;
	}
	std::cout << "Crossvalidation: " << (bCrossvalidate ? "yes" : "no") << std::endl;
		
	if(bNystroem && iMultiClassMode == 1)
	{
		iMultiClassMode = 0;
		std::cout << "iMultiClassMode 1 not supported for Nystroem approximated matrix " << std::endl;
	}
	            	
	if(bNystroem && iKernel !=2)
	{
		std::cout << " For Nystroem kernel switched to mode 2" << std::endl;
		iKernel = 2;
	}				    		
	
    }
    catch(...) {
        std::cerr << "Exception of unknown type - process with defaults!\n";
  }
	
	

  // Load the dataset.
  arma::mat DataT;
  bool bLoadSuccess = false;
  try 
  {
	  if(!bHDF5)
	  {
	  	bLoadSuccess = mlpack_stub::data::Load(sDataFilename, DataT, true,false);
	  }
	  else
	  {
		std::cout << "Load as hdf5 binary" << std::endl;
		bLoadSuccess = DataT.load(sDataFilename, hdf5_binary);
		//dataset = dataset.t();
	  }
  }
  catch(...) 
  {
        std::cerr << "Exception during file import!\n";
  }
 
  if(!bLoadSuccess)
  {
        std::cerr << "Loading input file failed!\n";
	return -1;
  }
  int M = DataT.n_rows;
  int N = DataT.n_cols;

  std::cout << "Have " << M << " rows and " << N << " columns" << std::endl;
	
  if(iKernel==2 & !bNystroem & M+2!=N & !bPredict)
  {
	std::cout << "kernel matrix needs to be squared (+2 columns index and label) " << std::endl;
	return -1;
  }

  arma::vec Label = vectorise(DataT.col(N-1));  
  
  if(iKernel == 2)
  {
	DataT = DataT.cols(1,N-2);
  }
  else
	DataT = DataT.cols(0,N-2);

  DataT.replace(datum::nan, 0.0); 

  boost::shared_ptr<mat> mSharedDataSet;
  cvm_parameters oParameters;
  oParameters.SetEpsilon(dEpsilon);
  oParameters.SetC(dC);
  oParameters.SetIterations (iCycles);
  if(bSparse)
	oParameters.SetSparsity(iSparse);
  else
	oParameters.SetSparsity(0);

  ICVMTrainInterface *oCVMTrain = NULL;
  IKernelMatrixInterface *oKernelMatrix = NULL;
  uvec vLandmarks;
  if(bIndefinite)
	oCVMTrain    = new icvm_train();
  else
	oCVMTrain    = new cvm_train();		

  if(bNystroem && iKernel == 2) // input is a Nystroem kernel)
  {
	oKernelMatrix = new NystroemKernelMatrix();
	
	vLandmarks = conv_to<uvec>::from(DataT.row(0));
	DataT      = DataT.rows(1,M-1);
	Label	   = Label.rows(1,M-1);
	M		   = M - 1;
	mSharedDataSet = boost::shared_ptr<mat>(new mat(DataT));
	boost::any oParam = vLandmarks;
	bResult = oKernelMatrix->Initialize(mSharedDataSet,oParam);				
  }
  else
  {
	mSharedDataSet= boost::shared_ptr<mat>(new mat(DataT));
	oKernelMatrix = new KernelMatrix (iKernel);
 	bResult = oKernelMatrix->Initialize(mSharedDataSet,zeros<mat>(1,1));	  
  }
  assert(bResult);
	
  oCVMTrain->SetParameters (oParameters );
  arma::vec vUniqueLabel = unique(Label);
	
  if(!bPredict)
  {
	  MultiClassModelFactory oMultiClassModelFactory((multimodeltype)iMultiClassMode);
	  boost::shared_ptr<IMultiClassModelInterface> sFullModel = oMultiClassModelFactory.GetMultiClassModel();// (new cvm_multiclass_model());
	  sFullModel->GenerateMultiClassModels(*oCVMTrain, *oKernelMatrix, DataT, Label);

	  // Save full model	 
	  if(!sModelFilename.empty())
	  {
		// Create an output archive

		std::ofstream ofs(sModelFilename.c_str());
		boost::archive::text_oarchive ar(ofs);
		//ofs.setf(std::ios_base::fixed, std::ios_base::floatfield);
		ofs.unsetf(std::ios_base::floatfield);
		// Write data
		ar & sFullModel;

		std::cout << "***" << std::endl;
		std::cout << sFullModel->GetUniqueLabels();
		std::cout << "***" << std::endl;
	  }	
	  
	  // predict the outcomes for the current full model
	  mat TrainingData;
	  boost::shared_ptr<mat> mSharedDataSetTraining;
	  if(iKernel != 2)
	  {
		  TrainingData = DataT.rows(join_cols(sFullModel->GetReferenceIndices(),linspace<ucolvec>(0,M-1,M)));
		  mSharedDataSetTraining = boost::shared_ptr<mat>(new mat(TrainingData));
		  std::cout << "* Used model indices:" << sFullModel->GetReferenceIndices().t() << std::endl;
	 	  bResult = oKernelMatrix->Initialize(mSharedDataSetTraining,zeros<mat>(1,1));

		  //std::cout << *mSharedDataSetTraining << std::endl;		  
	  }
	  else 
	  {
		  if(bNystroem)
		  {
			// for prediction we generate a full kernel reference-indices x data - so still roughly linear
			TrainingData = oKernelMatrix->Evaluate(sFullModel->GetReferenceIndices(),linspace<ucolvec>(0,M-1,M));
			 // std::cout << TrainingData << std::endl;
		    mSharedDataSetTraining = boost::shared_ptr<mat>(new mat(TrainingData));
		    std::cout << "* Used model indices:" << sFullModel->GetReferenceIndices().t() << std::endl;
			delete oKernelMatrix; // clear Nystroem instance
			oKernelMatrix = new KernelMatrix (2);
 			bResult = oKernelMatrix->Initialize(mSharedDataSetTraining,zeros<mat>(1,1));
		  }
		  else
		  {
			TrainingData = DataT.rows(sFullModel->GetReferenceIndices());
		    mSharedDataSetTraining = boost::shared_ptr<mat>(new mat(TrainingData));
		    std::cout << "* Used model indices:" << sFullModel->GetReferenceIndices().t() << std::endl;
 	 	    bResult = oKernelMatrix->Initialize(mSharedDataSetTraining,zeros<mat>(1,1));
		  }
	  }
	  assert(bResult);
	  mat PredictionResults = sFullModel->CalculateModelPredictions(*oKernelMatrix);
	  uvec vPredictedModel  = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));	  
	  std::cout << (double)sum(abs(sFullModel->GetUniqueLabels()(vPredictedModel)-Label)<std::numeric_limits<double>::epsilon())/(double)Label.size() << std::endl;
	  TrainingData.reset();
	  
	  if(bCrossvalidate)
	  {
			if(kFold>M)
			  kFold = M;
		  
			unsigned int iSetSize  = ceil(M/kFold);
			unsigned int iFullSize = kFold * iSetSize;
			uvec r = shuffle(linspace<uvec>(0,kFold-1,kFold));
			// V contains value between 0 and kFold-1 indicating whether an entry is included in the cv run or no
			uvec V = zeros<uvec>(M);
			uvec vShuffledIndices = shuffle(linspace<uvec>(0,M-1,M));
			if(M-iFullSize>0)
			{
				V(vShuffledIndices) =  join_vert(reshape(repmat(linspace<uvec>(0,kFold-1,kFold), iFullSize/kFold, 1), iFullSize, 1), r(linspace<uvec>(0,M-iFullSize-1,M-iFullSize)));
			}
			else
			{
				V(vShuffledIndices) =  reshape(repmat(linspace<uvec>(0,kFold-1,kFold), iFullSize/kFold, 1), iFullSize, 1);
			}

		  
			vec vPredictionAccuracy = zeros<vec>(kFold);

		    for(unsigned int k=0;k<kFold;++k)
			{
				// train model on training dat
				uvec vSelector = find(V!=k);

			    mat mSubsetData = DataT.rows(vSelector);			 
			  	if(iKernel == 2)
				{
					mSubsetData = mSubsetData.cols(vSelector);	
				}
				
	 			boost::shared_ptr<mat> mSharedDataSetCV = boost::shared_ptr<mat>(new mat(mSubsetData));
			 	bResult = oKernelMatrix->Initialize(mSharedDataSetCV,zeros<mat>(1,1));
				assert(bResult);
			
				
				vec SubLabel = Label(vSelector);

	  			MultiClassModelFactory oMultiClassModelFactoryCV((multimodeltype)iMultiClassMode);
				boost::shared_ptr<IMultiClassModelInterface> sFullModelCV = oMultiClassModelFactory.GetMultiClassModel();
				sFullModelCV->GenerateMultiClassModels(*oCVMTrain, *oKernelMatrix, mSubsetData, SubLabel);
				

				// make prediction on test data
				uvec vSelectorTest = find(V==k);
			  
	  			mat SubsetDataTest;
			    if(iKernel != 2)
			    {
					SubsetDataTest = DataT.rows(join_cols(vSelector.rows(sFullModel->GetReferenceIndices()),vSelectorTest));
				}
				else
			    {
					SubsetDataTest = DataT.rows(vSelector.rows(sFullModel->GetReferenceIndices()));
					SubsetDataTest = SubsetDataTest.cols(vSelectorTest);
				}
	 			boost::shared_ptr<mat> mSharedDataSetTest = boost::shared_ptr<mat>(new mat(SubsetDataTest));
				bResult = oKernelMatrix->Initialize(mSharedDataSetTest,zeros<mat>(1,1));
		  
				assert(bResult);
				mat PredictionResults= sFullModelCV->CalculateModelPredictions(*oKernelMatrix);  			
			        uvec vPredictedModel = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
				vPredictionAccuracy(k) = (double)sum(abs(sFullModelCV->GetUniqueLabels()(vPredictedModel)-Label(vSelectorTest))<std::numeric_limits<double>::epsilon())/(double)vSelectorTest.size();
			}
			std::cout << mean(vPredictionAccuracy) << " " << sqrt(var(vPredictionAccuracy)) << std::endl; 
	  }
  }
  else if(!sModelFilename.empty())
  {
	  // Restore data
	  std::cout << "restore ..." << std::endl;
	  boost::shared_ptr<IMultiClassModelInterface> sFullModelRec;//( new cvm_multiclass_model());
	  {
		// Create and input archive
		std::ifstream ifs(sModelFilename.c_str());
		ifs.setf(std::ios_base::fixed, std::ios_base::floatfield);
		boost::archive::text_iarchive ar(ifs);
		ar & sFullModelRec ;
		  std::cout << "restore complete ..." << std::endl;


		cvm_model oModel1 = sFullModelRec->GetModels()[0]; // just choose the first model (we will not support different kernels on multiclass models but take always the same
		KernelMatrix oKernelMatrix(oModel1.GetKernelType());
		bool bResult = false;

		// verify if input data fit to the model
		arma::uvec vReferenceIndices = sFullModelRec->GetReferenceIndices();
		bool bInputDataValid = false;
		if(iKernel < 2) //vectorial data
		{	
			if(sFullModelRec->GetReferenceData()[0].size() == DataT.n_cols)
			{
				bInputDataValid = true;
			}
		}
		else
		{
			if(vReferenceIndices.n_elem == DataT.n_rows)
			{
				bInputDataValid = true;
			}

		}

		if(!bInputDataValid)
		{
			std::cerr << "Invalid test data check model" << std::endl;
			return -1;
		}

		if(iKernel < 2)
		{		  
			const std::vector< std::vector<double> >& vvReferenceData = sFullModelRec->GetReferenceData();
			unsigned int NPoints = vvReferenceData.size();
			mat ReferenceData = zeros<mat>(NPoints,vvReferenceData[0].size());
			for(unsigned int k=0; k<NPoints; k++)
			{				
				arma::vec vTemp = arma::conv_to< arma::vec >::from(vvReferenceData[k]);
				ReferenceData.row(k)  = vTemp.t();
			}		  
			mat mReferenceAndTestData = join_cols(ReferenceData,DataT);
			boost::shared_ptr<mat> mSharedReferenceAndTestData = boost::shared_ptr<mat>(new mat(mReferenceAndTestData));
			bResult = oKernelMatrix.Initialize(mSharedReferenceAndTestData,zeros<mat>(1,1));
		}
		else
		{
			boost::shared_ptr<mat> mSharedReferenceAndTestData = boost::shared_ptr<mat>(new mat(DataT));
			oKernelMatrix = KernelMatrix (2);
			bResult = oKernelMatrix.Initialize(mSharedReferenceAndTestData,zeros<mat>(1,1));
		}
			  
	  	std::cout << "start prediction ..." << std::endl;
		assert(bResult);
		  
		mat PredictionResults    = sFullModelRec->CalculateModelPredictions(oKernelMatrix);
		vec vNormalizer          = sum(PredictionResults.cols(0,PredictionResults.n_cols-2),1);
		PredictionResults.cols(0,PredictionResults.n_cols-2)    	 = PredictionResults.cols(0,PredictionResults.n_cols-2) / repmat(vNormalizer,1,PredictionResults.n_cols-1);

		uvec vPredictedClassIndex    = conv_to< uvec >::from(PredictionResults.col(PredictionResults.n_cols-1));
  		vec vPredictedClassLabel = sFullModelRec->GetUniqueLabels()(vPredictedClassIndex);		  
//		std::cout << PredictionResults << std::endl;
//		std::cout << sum(PredictionResults.t()) << std::endl;
		std::cout << vPredictedClassLabel << std::endl;
	  }
  }
  else
  {
		std::cout << "Need to specify a model file for prediction" << std::endl;
  }

  delete oCVMTrain;
  delete oKernelMatrix;	
  return 0;
}
