/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * EIG_NY_IMPL.hpp
 * Copyright (C) 2015 Frank-Michael Schleif <schleify@cs.bham.ac.uk>
 *
 * This is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Implementation of an elm kernel in acc to
 * Frenay & Verleysen, Parameter Free kernel in extreme learnin ...,
 * Neurocomputing
 * Author: Frank-Michael Schleif
 */
#ifndef _EIG_NY_IMPL_HPP_
#define _EIG_NY_IMPL_HPP_

#include <iostream>
#include <armadillo>
#include "eig_ny.hpp"
#include <cmath>

using namespace std;
using namespace arma;

/*
% computes the exact eigenvalue decomposition of a Nystroem approximated
% matrix
% Input agruments:
%   ny_K is a Nystroem approximated matrix of form
%     ny_K{1}=K_Nm
%     ny_K{2}=K_mm^-1
% Output agruments:
%   C is the matrix of eigenvectors
%   A is the diagonal matrix of eigenvalues
*/

void eig_ny(const mat &C, const mat &Wpinv, mat &E, mat &V)
{
	// get eigenvectors of approximated matrix
	mat Temp = Wpinv*(C.t()*C)*Wpinv;
	Temp = 0.5 * (Temp+Temp.t());

	vec eigval;
	mat eigvec;

	eig_sym(eigval, eigvec, Temp);
	Temp.reset();
	
	uvec vInvalid = find(eigval < std::numeric_limits<double>::epsilon());
	eigval.elem( vInvalid ).zeros();

	mat B=C*(eigvec*sqrt(diagmat(eigval)));     // B = is the left / right part of the K_Nm * K_mm^-1 * K_Nm with inserted decomposition of K_mm^-1 in eigenvalues (U) and eigenvectors (L) 

	eigval.reset();
	eigvec.reset();
	
	vec eigvalA;
	mat eigvecV;
	eig_sym(eigvalA,eigvecV,B.t()*B);        // eigenvalue decomposition of the covariance matrix of B to get eigenvectors V and eigenvalue matrix A

	vInvalid = find(eigvalA < std::numeric_limits<double>::epsilon());
	eigvalA.elem(vInvalid).zeros();
	E=B*(eigvecV*sqrt(pinv(diagmat(eigvalA)))); // now we have the eigenvector C of K and actually already the eigenvalue matrix A real(diag(sqrt(A))) - this works for psd matrices

	eigvecV.reset();
	eigvalA.reset();
	B.reset();
/*
% get eigenvalues of approximated matrix - for non-psd we have a
% to reconstruct the eigenvalues A as follows to avoid the sqrt(-1) issue
% Eigenvalue equation Av = lv
% get eigenvalues by starting with the eigenvectors C of A
% l  = v^-1 * A * v
% l  = (v^-1 * ny_K{1}) * ny_K{2} * ( ny_K{1} *v)
*/

//	[A,idx]=sort(diag((C'*ny_K{1})*ny_K{2}*(ny_K{1}'*C)),'descend');
	vec  vTemp   = diagvec((E.t()*C)*Wpinv*(C.t()*E));

	uvec indices = sort_index(vTemp,1);
	E            = E.cols(indices);
	V            = diagmat(vTemp(indices));
}
#endif
