/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * cvmtrain.cc
 * Copyright (C) 2015 Frank-Michael Schleif <schleify@cs.bham.ac.uk>
 *
 * This is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cvmtrain.h"
#include "iqp_solver.h"
#include "math_stat_helper.hpp"
#include "ompursuit.h"
#include <cmath>

using namespace arma;
cvm_train::cvm_train()
{
	m_oParameters = cvm_parameters ();
	m_oCVMModel  = cvm_model (zeros<vec>(1,1),0.0); // dummy model
}

cvm_train::~cvm_train()
{
}

void cvm_train::SetParameters(cvm_parameters oParameters)
{
	m_oParameters = oParameters;
}

cvm_model cvm_train::GetModel(void)
{
	return m_oCVMModel;	
}


void cvm_train::Apply(const IKernelMatrixInterface &oKernel,const vec &Y)
{
	// the following is a c++ implementation of the Core Vector Machine by I. Tsang

	// the maximal iterations - just a backup
	unsigned int niters = this->m_oParameters.GetIterations ();
	unsigned int rows   = 0;
	unsigned int cols   = 0;
	oKernel.size(rows,cols);
	unsigned int N      = rows;

	vec w;

	// Initial weight vector to let w to be large than zero
	w = zeros<vec>(N,1);

	// dEpsilon
	double dEpsilon = this->m_oParameters.GetEpsilon ();

	// parameter C - upper error bound
	double dC       = this->m_oParameters.GetC ();	

	colvec vDiags = zeros<vec>(N);
	
	for(unsigned int i=0; i!=N;++i)
	{
		vDiags(i) = as_scalar(oKernel.Evaluate(uvec({i}), uvec({i})));
	}	

	ucolvec S;
	double  R = 0;

	colvec vAlpha;	
	colvec vDelta= (max(vDiags)-vDiags);

	double iResult   = CoreSetSolver(oKernel, Y, vDiags, dEpsilon, vDelta, S, R, vAlpha);
	double dBias     = as_scalar(vAlpha.t()*Y(S));
/*
	std::cout << S.t() << " : " << R << std::endl; 
	std::cout << vAlpha.t() << std::endl;
	std::cout << dBias << std::endl;
	*/
	w.rows(S)   = vAlpha; 
	m_oCVMModel = cvm_model(w % Y,dBias,oKernel.GetKernelType()); 
}

cvm_parameters cvm_train::GetParameters(void)
{
	return m_oParameters;
}


//int cvm_train::CoreSetSolver(const mat &K, double epsilon,const colvec &vDiags, colvec &vDeltas, ucolvec &S, double &R, colvec &vAlpha)
int cvm_train::CoreSetSolver(const IKernelMatrixInterface &oKernel,const vec &Y, const colvec &vDiagsK, double epsilon,  colvec &vDeltas, ucolvec &S, double &R, colvec &vAlpha)
{
	int    iResult = 0;
	int    N          = vDiagsK.n_elem;
	double dTolerance = (1+epsilon)*(1+epsilon);
	double dC         = this->m_oParameters.GetC ();
	
	// default init
	ucolvec Sdefault = shuffle(linspace<ucolvec>(0,N-1,N));
	S                = Sdefault(uvec({0,1}));//randi<ucolvec>(2, distr_param(0,N-1) );    // choose 2 random points
	Sdefault         = Sdefault(find(Sdefault!=S(0))); // remove used entries
	Sdefault         = Sdefault(find(Sdefault!=S(1)));

//	std::cout << Sdefault.n_elem << std::endl;
	ucolvec SLast= S;
    vAlpha       = 0.5*ones<colvec>(2);                      // default loadings in case of 2 points (initial center)
	colvec vAlphaLast = vAlpha;
	 
	//mat K     = oKernel.GetKernelMatrix ();;
	//mat mDiagC= diagmat(ones<vec>(N,1)/dC);
	//mat mYY   = Y*Y.t();
	//mat K2CVM = K % mYY +  mYY + mDiagC;
	//vec vDiags    = K2CVM.diag();
	//K            = K2CVM;
	const colvec vDiags = vDiagsK+2.0/dC; 
	
	mat mKSS     = oKernel.Evaluate(S,S) % (Y(S)*Y(S).t()) + Y(S)*Y(S).t() + eye<mat>(S.n_elem,S.n_elem)/dC;   //K(S,S);

	R            = as_scalar((vAlpha.t() * (mKSS.diag()+vDeltas(S)) -vAlpha.t() * mKSS * vAlpha));   // thats R^2 - initial radius
	
	int iProbSampleSize          = 59;
	int NDefault                 = Sdefault.n_elem;
	ucolvec vDefaultRandSampling = shuffle(Sdefault);	
	ucolvec vRandSampling        = vDefaultRandSampling(linspace<ucolvec>(0,NDefault-1,min(iProbSampleSize,NDefault) ));//randi<ucolvec>(min(iProbSampleSize,N-1), distr_param(0,N-1) );

	mat mZ       = zeros<mat>(S.n_elem,vRandSampling.n_elem);
	for(int k=0;k != S.n_elem; ++k)
		for(int i=0; i != vRandSampling.n_elem; ++i)
			mZ(k,i) = (double)(S(k)==vRandSampling(i));
	mZ = mZ/dC;
	//mat mK       = oKernel.Evaluate(S,vRandSampling) % (Y(S)*Y(vRandSampling).t()) + Y(S)*Y(vRandSampling).t() + mDiagC(S,vRandSampling);  //K(S,vRandSampling);  // kernel of core points to all out of radius points
	mat mK       = oKernel.Evaluate(S,vRandSampling) % (Y(S)*Y(vRandSampling).t()) + Y(S)*Y(vRandSampling).t() + mZ;  //K(S,vRandSampling);  // kernel of core points to all out of radius points
	//vec vDistances = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + ones<vec>(vRandSampling.n_elem).t()*as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
	vec vDistances = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
	uvec vOutOfRadius = find(vDistances > dTolerance * R);

	// loop  until MEB is found
	int iCounter     = 2;
	double Rold         = R;
	iqpsolver	   iqpSolverInstance;
	while(vOutOfRadius.n_elem>0)
	{		
		uword I;
		vDistances(vOutOfRadius).max(I);                         	   // largest point away - original core-set heuristic
        uword iNewCoreIndex = vRandSampling(vOutOfRadius(I));    	   // identify new coreset point
		Sdefault            = Sdefault(find(Sdefault!=iNewCoreIndex)); // remove new added entry
//		std::cout << Sdefault.n_elem << "  " << iNewCoreIndex << std::endl;
/*		std::cout << "*****------*****  " << R << std::endl;
		std::cout << vDistances(vOutOfRadius) << std::endl;
		std::cout << vOutOfRadius<< std::endl;
		std::cout << "*****------*****" << std::endl;
		std::cout << iNewCoreIndex << " - " << vDistances(vOutOfRadius(I)) << std::endl;
*/
		
		iCounter         = iCounter+1;
        SLast            = S;
		S				 = join_cols<umat>(S, uvec({iNewCoreIndex}));  // add to core-set		
        mKSS             = oKernel.Evaluate(S,S) % (Y(S)*Y(S).t()) + Y(S)*Y(S).t() + eye<mat>(S.n_elem,S.n_elem)/dC; // K(S,S);
        vAlphaLast       = vAlpha;
        int Ns           = S.n_elem;
        colvec f         = zeros<vec>(Ns); 
        vec ones1N       = ones<vec>(Ns);
        colvec zerosN1   = zeros<colvec>(Ns);
        colvec onesN1    = ones<colvec>(Ns);
        mKSS             = (mKSS + mKSS.t()) / 2.0;
		double dVal      = 0;

		mat fullA	     = diagmat(ones<vec>(Ns));				// alpha>=0
		fullA	         = join_cols(fullA, ones<vec>(Ns).t()); // sum alpha <= 1
		fullA	         = join_cols(fullA, ones<vec>(Ns).t()); // sum alpha >= 1
		colvec fullb     = zerosN1;								// alpha>=0		
	    fullb            = join_cols<mat>(fullb, vec({0}));		// sum alpha <= 1
		fullb            = join_cols<mat>(fullb, vec({-1}));	// sum alpha >= 1

/*		std::cout << ">>>>" << std::endl;
		std::cout << mKSS << " - " << f <<  " - " << fullA << " - " << fullb << " - " << vAlpha << std::endl;
		std::cout << ">>>>" << std::endl;
	*/	
		iResult          = iqpSolverInstance.InteriorPointSolver(2*mKSS, f, fullA, fullb, true, vAlpha, dVal);
		R                = as_scalar((vAlpha.t() * (mKSS.diag()+vDeltas(S)) -vAlpha.t() * mKSS * vAlpha));   // thats R^2 - initial radius
		vAlpha           = vAlpha/sum(vAlpha);
		/*cin.get();
		std::cout << vAlpha << " - " << R << std::endl;*/

		NDefault                     = Sdefault.n_elem;
		ucolvec vDefaultRandSampling = shuffle(Sdefault);	
		vRandSampling                = vDefaultRandSampling(linspace<ucolvec>(0,NDefault-1,min(iProbSampleSize,NDefault) ));//randi<ucolvec>(min(iProbSampleSize,N-1), distr_param(0,N-1) );

		mat mZ       = zeros<mat>(S.n_elem,vRandSampling.n_elem);
		for(int k=0;k != S.n_elem; ++k)
			for(int i=0; i != vRandSampling.n_elem; ++i)
				mZ(k,i) = (double)(S(k)==vRandSampling(i));
		mZ = mZ/dC;
		//mK               = oKernel.Evaluate(S,vRandSampling) % (Y(S)*Y(vRandSampling).t()) + Y(S)*Y(vRandSampling).t() + mDiagC(S,vRandSampling);// K(S,vRandSampling);
		mK               = oKernel.Evaluate(S,vRandSampling) % (Y(S)*Y(vRandSampling).t()) + Y(S)*Y(vRandSampling).t() + mZ;// K(S,vRandSampling);
		vDistances       = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
		vOutOfRadius     = find(vDistances > dTolerance * R);
		//std::cout << sort(Sdefault.t()) << std::endl;
		//std::cout << "vR: " << vRandSampling.t() << std::endl;
		//std::cout << vOutOfRadius.n_elem << std::endl;
	}
	return 0;
}

/*
void cvm_train::Apply(const IKernelMatrixInterface &oKernel,const vec &Label)
{
	// the following is a c++ implementation of the Core Vector Machine by I. Tsang

	// the maximal iterations - just a backup
	const mat &Data = oKernel.GetKernelMatrix ();
	unsigned int niters = this->m_oParameters.GetIterations ();
	unsigned int ndata  = Data.n_rows;

	vec w;

	// Initial weight vector to let w to be large than zero
	w = zeros<vec>(ndata,1);

	// dEpsilon
	double dEpsilon = this->m_oParameters.GetEpsilon ();

	// parameter C - upper error bound
	double dC       = this->m_oParameters.GetC ();

	vec Y = Label;

	mat K = Data;

	colvec vDiags   = K.diag();
	int    N        = vDiags.n_elem;	

	ucolvec S;
	double  R = 0;

	colvec vAlpha;	
	colvec vDelta= (max(vDiags)-vDiags);

	mat mDiagC= diagmat(ones<vec>(N,1)/dC);
	mat mYY   = Y*Y.t();
	mat K2CVM = K % mYY +  mYY + mDiagC;
	vDiags    = K2CVM.diag();
	double iResult   = CoreSetSolver(K2CVM, dEpsilon,vDiags, vDelta, S, R, vAlpha);
	double dBias     = as_scalar(vAlpha.t()*Y(S));
	w.rows(S)   = vAlpha; 
	m_oCVMModel = cvm_model(w % Label,dBias,oKernel.GetKernelType()); 
}

cvm_parameters cvm_train::GetParameters(void)
{
	return m_oParameters;
}


int cvm_train::CoreSetSolver(const mat &K, double epsilon,const colvec &vDiags, colvec &vDeltas, ucolvec &S, double &R, colvec &vAlpha) 
{
	int    iResult = 0;
	int    N          = vDiags.n_elem;
	double dTolerance = (1+epsilon)*(1+epsilon);
	
	// default init
	ucolvec Sdefault = shuffle(linspace<ucolvec>(0,N-1,N));
	S                = Sdefault(uvec({0,1}));//randi<ucolvec>(2, distr_param(0,N-1) );    // choose 2 random points
	Sdefault         = Sdefault(find(Sdefault!=S(0))); // remove used entries
	Sdefault         = Sdefault(find(Sdefault!=S(1)));

//	std::cout << Sdefault.n_elem << std::endl;
	ucolvec SLast= S;
    vAlpha       = 0.5*ones<colvec>(2);                      // default loadings in case of 2 points (initial center)
	colvec vAlphaLast = vAlpha;
	mat mKSS     = K(S,S);

	R            = as_scalar((vAlpha.t() * (mKSS.diag()+vDeltas(S)) -vAlpha.t() * mKSS * vAlpha));   // thats R^2 - initial radius
	
	int iProbSampleSize          = 59;
	int NDefault                 = Sdefault.n_elem;
	ucolvec vDefaultRandSampling = shuffle(Sdefault);	
	ucolvec vRandSampling        = vDefaultRandSampling(linspace<ucolvec>(0,NDefault-1,min(iProbSampleSize,NDefault) ));//randi<ucolvec>(min(iProbSampleSize,N-1), distr_param(0,N-1) );
	
	mat mK       = K(S,vRandSampling);  // kernel of core points to all out of radius points
	mKSS         = K(S,S);  			// SFull is without S

	//vec vDistances = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + ones<vec>(vRandSampling.n_elem).t()*as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
	vec vDistances = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
	uvec vOutOfRadius = find(vDistances > dTolerance * R);

	// loop  until MEB is found
	int iCounter     = 2;
	double Rold         = R;
	iqpsolver	   iqpSolverInstance;
	while(vOutOfRadius.n_elem>0)
	{		
		uword I;
		vDistances(vOutOfRadius).max(I);                         	   // largest point away - original core-set heuristic
        uword iNewCoreIndex = vRandSampling(vOutOfRadius(I));    	   // identify new coreset point
		Sdefault            = Sdefault(find(Sdefault!=iNewCoreIndex)); // remove new added entry
//		std::cout << Sdefault.n_elem << "  " << iNewCoreIndex << std::endl;
		
		iCounter         = iCounter+1;
        SLast            = S;
		S				 = join_cols<umat>(S, uvec({iNewCoreIndex}));  // add to core-set		
        mKSS             = K(S,S);
        vAlphaLast       = vAlpha;
        int Ns           = S.n_elem;
        colvec f         = zeros<vec>(Ns); 
        vec ones1N       = ones<vec>(Ns);
        colvec zerosN1   = zeros<colvec>(Ns);
        colvec onesN1    = ones<colvec>(Ns);
        mKSS             = (mKSS + mKSS.t()) / 2.0;
		double dVal      = 0;

		mat fullA	     = diagmat(ones<vec>(Ns));				// alpha>=0
		fullA	         = join_cols(fullA, ones<vec>(Ns).t()); // sum alpha <= 1
		fullA	         = join_cols(fullA, ones<vec>(Ns).t()); // sum alpha >= 1
		colvec fullb     = zerosN1;								// alpha>=0		
	    fullb            = join_cols<mat>(fullb, vec({0}));		// sum alpha <= 1
		fullb            = join_cols<mat>(fullb, vec({-1}));	// sum alpha >= 1

		iResult          = iqpSolverInstance.InteriorPointSolver(2*mKSS, f, fullA, fullb, true, vAlpha, dVal);
		R                = as_scalar((vAlpha.t() * (mKSS.diag()+vDeltas(S)) -vAlpha.t() * mKSS * vAlpha));   // thats R^2 - initial radius
		vAlpha           = vAlpha/sum(vAlpha);

		NDefault                     = Sdefault.n_elem;
		ucolvec vDefaultRandSampling = shuffle(Sdefault);	
		vRandSampling                = vDefaultRandSampling(linspace<ucolvec>(0,NDefault-1,min(iProbSampleSize,NDefault) ));//randi<ucolvec>(min(iProbSampleSize,N-1), distr_param(0,N-1) );

		mK               = K(S,vRandSampling);
		vDistances       = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
		vOutOfRadius     = find(vDistances > dTolerance * R);
		//std::cout << sort(Sdefault.t()) << std::endl;
		//std::cout << "vR: " << vRandSampling.t() << std::endl;
		//std::cout << vOutOfRadius.n_elem << std::endl;
	}
	mK               = K.rows(S);
	vDistances       = (vDiags.t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas.t() ).t(); // get distances for MEB
	vOutOfRadius     = find(vDistances > dTolerance * R);
//	std::cout << vDistances(vOutOfRadius) << " " << dTolerance * R << std::endl;
	return 0;
}

*/ // batch code
