/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * icvmtrain.cc
 * Copyright (C) 2015 Frank-Michael Schleif <schleify@cs.bham.ac.uk>
 *
 * This is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "icvmtrain.h"
#include "iqp_solver.h"
#include "eig_ny.hpp"
#include "ompursuit.h"
#include "math_stat_helper.hpp"
#include "nystroem-kernel-matrix.h"
#include <boost/shared_ptr.hpp>
#include <cmath>

using namespace arma;
icvm_train::icvm_train()
{
	m_oParameters = cvm_parameters ();
	m_oCVMModel  = cvm_model (zeros<vec>(1,1),0.0); // dummy model
}

icvm_train::~icvm_train()
{
}

void icvm_train::SetParameters(cvm_parameters oParameters)
{
	m_oParameters = oParameters;
}

cvm_model icvm_train::GetModel(void)
{
	return m_oCVMModel;	
}


void icvm_train::Apply(const IKernelMatrixInterface &oKernel,const vec &Y)
{
	// the following is a c++ implementation of the Core Vector Machine by I. Tsang

	// the maximal iterations - just a backup
	unsigned int rows   = 0;
	unsigned int cols   = 0;
	oKernel.size(rows,cols);
	unsigned int N      = rows;
	
	vec w;

	// Initial weight vector to let w to be large than zero
	w = zeros<vec>(N,1);

	// dEpsilon - is chosen much smaller for indefinite kernels to be more sensitive
	double dEpsilon = 0.001*this->m_oParameters.GetEpsilon ();

	// parameter C - upper error bound
	double dC       = this->m_oParameters.GetC ();	
	
	colvec vDiags = zeros<vec>(N);
	
	for(unsigned int i=0; i!=N;++i)
	{
		vDiags(i) = as_scalar(oKernel.Evaluate(uvec({i}), uvec({i})));
	}	

	// preprocessing to handle indefinite input kernels
	ucolvec vAllIndices= linspace<ucolvec>(0,N-1,N);
	ucolvec Sdefault   = shuffle(vAllIndices);
	ucolvec vLandmarks;
	if(N>2000)
		vLandmarks = Sdefault(uvec({0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29})); //choose 10 random indices
	else
		vLandmarks = Sdefault;
	mat mTemp = zeros<mat>(N,vLandmarks.n_elem);
	mTemp.rows(vLandmarks) = eye<mat>(vLandmarks.n_elem,vLandmarks.n_elem)/dC;
	mat nyC   = oKernel.Evaluate(vAllIndices, vLandmarks) % (Y*Y(vLandmarks).t()) + Y*Y(vLandmarks).t() + mTemp;
	mat nyW   = pinv(nyC.rows(vLandmarks));
		
	mat eC, vA;
	eig_ny(nyC, nyW, eC, vA);
	
	uvec vNeg       = find(sign(vA.diag())<0);
	mat ny_Kabs     = eC * abs(vA) * eC.rows(vLandmarks).t();
	mat ny_KabsW    = ny_Kabs.rows(vLandmarks);
	ny_KabsW        = 0.5 * (ny_KabsW+ny_KabsW.t());
	ny_KabsW        = pinv(ny_KabsW);

	IKernelMatrixInterface *oKernelMatrix = NULL;
	boost::shared_ptr<mat> mSharedDataSet;

	oKernelMatrix = new NystroemKernelMatrix();
	mSharedDataSet = boost::shared_ptr<mat>(new mat(ny_Kabs));
	boost::any oParam = vLandmarks;
	oKernelMatrix->Initialize(mSharedDataSet,oParam);

	
	vDiags         = sum((ny_Kabs * ny_KabsW) % ny_Kabs,1); // approximated diagonal elements    
	
	ucolvec S;
	double  R = 0;

	colvec vAlpha;	
	colvec vDelta= (max(vDiags)-vDiags);

	double iResult   = CoreSetSolver(*oKernelMatrix, Y, vDiags, dEpsilon, vDelta, S, R, vAlpha);

	std::cout << "Initial complexity " << S.n_elem <<std::endl;
    vec vAlphaMod        = eC.rows(S).t()*vAlpha;
    vAlphaMod.rows(vNeg) = -vAlphaMod.rows(vNeg);
    vAlphaMod	         = (eC*vAlphaMod);

	int iSparsity = this->m_oParameters.GetSparsity();
/*
	std::cout << S.t() << " : " << R << std::endl; 
	std::cout << vAlpha.t() << std::endl;
	std::cout << dBias << std::endl;
	*/

	if(iSparsity >  vLandmarks.n_elem)
	{
		std::cout << "Warning: Sparsity reset to max landmark size" << std::endl;
		iSparsity = vLandmarks.n_elem;
	}

	if(iSparsity > 0)
	{
		omp oOmp;
		unsigned int iSparseDegree = iSparsity;
	//	std::cout << vAlphaMod << std::endl;

	//	mat      A  = oKernelMatrix->Evaluate(vAllIndices,vAllIndices);
	
	/*	KernelMatrix oKernelTemp = KernelMatrix(2);
		boost::shared_ptr<mat> mSharedDataSet2 = boost::shared_ptr<mat>(new mat(A));
		oKernelTemp.Initialize(mSharedDataSet2,zeros<mat>(1,1));
	*/
	//	vec      y  = A * vAlphaMod;
		const mat& C = (dynamic_cast<const NystroemKernelMatrix*>(oKernelMatrix))->GetParameterC();
		const mat& W = (dynamic_cast<const NystroemKernelMatrix*>(oKernelMatrix))->GetParameterP();
		vec      y = C*((C.t()*vAlphaMod).t()*W).t();
		vAlphaMod   = oOmp.OmpSolver(*oKernelMatrix, y, iSparseDegree);
	//    vAlphaMod   = oOmp.OmpSolver(oKernelTemp, y, iSparsity);
	}
	w           = vAlphaMod; 
	//std::cout << "Sum w: " << sum(w) << std::endl;
	// bias remains always the same in sparse and non-sparse representation
	double dBias         = as_scalar(vAlphaMod.t() * Y);
	m_oCVMModel = cvm_model(w % Y,dBias,oKernel.GetKernelType()); 
}

cvm_parameters icvm_train::GetParameters(void)
{
	return m_oParameters;
}


//int icvm_train::CoreSetSolver(const mat &K, double epsilon,const colvec &vDiags, colvec &vDeltas, ucolvec &S, double &R, colvec &vAlpha)
int icvm_train::CoreSetSolver(IKernelMatrixInterface &oKernel,const vec &Y, const colvec &vDiags, double epsilon,  colvec &vDeltas, ucolvec &S, double &R, colvec &vAlpha)
{
	int    iResult = 0;
	unsigned int niters = this->m_oParameters.GetIterations ();
	int    N          = vDiags.n_elem;
	double dTolerance = (1+epsilon)*(1+epsilon);
	double dC         = this->m_oParameters.GetC ();
	
	// default init
	ucolvec Sdefault = shuffle(linspace<ucolvec>(0,N-1,N));
	S                = Sdefault(uvec({0,1}));//randi<ucolvec>(2, distr_param(0,N-1) );    // choose 2 random points
	Sdefault         = Sdefault(find(Sdefault!=S(0))); // remove used entries
	Sdefault         = Sdefault(find(Sdefault!=S(1)));

//	std::cout << Sdefault.n_elem << std::endl;
	ucolvec SLast= S;
    vAlpha       = 0.5*ones<colvec>(2);                      // default loadings in case of 2 points (initial center)
	colvec vAlphaLast = vAlpha;
	 
	//mat K     = oKernel.GetKernelMatrix ();;
	//mat mDiagC= diagmat(ones<vec>(N,1)/dC);
	//mat mYY   = Y*Y.t();
	//mat K2CVM = K % mYY +  mYY + mDiagC;
	//vec vDiags    = K2CVM.diag();
	//K            = K2CVM;
	
	mat mKSS     = oKernel.Evaluate(S,S);   //K(S,S);

	R            = as_scalar((vAlpha.t() * (mKSS.diag()+vDeltas(S)) -vAlpha.t() * mKSS * vAlpha));   // thats R^2 - initial radius
	
	int iProbSampleSize          = 59;
	int NDefault                 = Sdefault.n_elem;
	ucolvec vDefaultRandSampling = shuffle(Sdefault);	
	ucolvec vRandSampling        = vDefaultRandSampling(linspace<ucolvec>(0,NDefault-1,min(iProbSampleSize,NDefault) ));//randi<ucolvec>(min(iProbSampleSize,N-1), distr_param(0,N-1) );

	mat mK            = oKernel.Evaluate(S,vRandSampling);  //K(S,vRandSampling);  // kernel of core points to all out of radius points
	vec vDistances    = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
	uvec vOutOfRadius = find(vDistances > dTolerance * R);

	// loop  until MEB is found
	int iCounter     = 2;
	double Rold         = R;
	iqpsolver	   iqpSolverInstance;
	while(vOutOfRadius.n_elem>0)
	{		
		uword I;
		vDistances(vOutOfRadius).max(I);                         	   // largest point away - original core-set heuristic
        uword iNewCoreIndex = vRandSampling(vOutOfRadius(I));    	   // identify new coreset point
		Sdefault            = Sdefault(find(Sdefault!=iNewCoreIndex)); // remove new added entry
//		std::cout << Sdefault.n_elem << "  " << iNewCoreIndex << std::endl;
/*		std::cout << "*****------*****  " << R << std::endl;
		std::cout << vDistances(vOutOfRadius) << std::endl;
		std::cout << vOutOfRadius<< std::endl;
		std::cout << "*****------*****" << std::endl;
		std::cout << iNewCoreIndex << " - " << vDistances(vOutOfRadius(I)) << std::endl;
*/
		
		iCounter         = iCounter+1;
        SLast            = S;
		S				 = join_cols<umat>(S, uvec({iNewCoreIndex}));  // add to core-set		
        mKSS             = oKernel.Evaluate(S,S); // K(S,S);
        vAlphaLast       = vAlpha;
        int Ns           = S.n_elem;
        colvec f         = zeros<vec>(Ns); 
        vec ones1N       = ones<vec>(Ns);
        colvec zerosN1   = zeros<colvec>(Ns);
        colvec onesN1    = ones<colvec>(Ns);
        mKSS             = (mKSS + mKSS.t()) / 2.0;
		double dVal      = 0;

		mat fullA	     = diagmat(ones<vec>(Ns));				// alpha>=0
		fullA	         = join_cols(fullA, ones<vec>(Ns).t()); // sum alpha <= 1
		fullA	         = join_cols(fullA, ones<vec>(Ns).t()); // sum alpha >= 1
		colvec fullb     = zerosN1;								// alpha>=0		
	    fullb            = join_cols<mat>(fullb, vec({0}));		// sum alpha <= 1
		fullb            = join_cols<mat>(fullb, vec({-1}));	// sum alpha >= 1

/*		std::cout << ">>>>" << std::endl;
		std::cout << mKSS << " - " << f <<  " - " << fullA << " - " << fullb << " - " << vAlpha << std::endl;
		std::cout << ">>>>" << std::endl;
	*/	
		//std::cout << S.t() << std::endl;
		//std::cout << mKSS << std::endl;
		iResult          = iqpSolverInstance.InteriorPointSolver(2*mKSS, f, fullA, fullb, true, vAlpha, dVal);
		R                = as_scalar((vAlpha.t() * (mKSS.diag()+vDeltas(S)) -vAlpha.t() * mKSS * vAlpha));   // thats R^2 - initial radius
		vAlpha           = vAlpha/sum(vAlpha);
		/*cin.get();
		std::cout << vAlpha << " - " << R << std::endl;*/

		NDefault                     = Sdefault.n_elem;
		if(NDefault == 0)
			return 0;
		ucolvec vDefaultRandSampling = shuffle(Sdefault);	
		vRandSampling                = vDefaultRandSampling(linspace<ucolvec>(0,NDefault-1,min(iProbSampleSize,NDefault) ));//randi<ucolvec>(min(iProbSampleSize,N-1), distr_param(0,N-1) );
		
		mK               = oKernel.Evaluate(S,vRandSampling);// K(S,vRandSampling);
		vDistances       = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
		vOutOfRadius     = find(vDistances > dTolerance * R);
		//std::cout << sort(Sdefault.t()) << std::endl;
		//std::cout << "vR: " << vRandSampling.t() << std::endl;
		//std::cout << vOutOfRadius.n_elem << std::endl;

		if(iCounter > niters)
		{
			std::cerr << " MEB calculation reached iteration limit - stop with approximate result" << std::endl;
			return -1;
		}
	}
	return 0;
}

/*
void icvm_train::Apply(const IKernelMatrixInterface &oKernel,const vec &Label)
{
	// the following is a c++ implementation of the Core Vector Machine by I. Tsang

	// the maximal iterations - just a backup
	const mat &Data = oKernel.GetKernelMatrix ();
	unsigned int niters = this->m_oParameters.GetIterations ();
	unsigned int ndata  = Data.n_rows;

	vec w;

	// Initial weight vector to let w to be large than zero
	w = zeros<vec>(ndata,1);

	// dEpsilon
	double dEpsilon = this->m_oParameters.GetEpsilon ();

	// parameter C - upper error bound
	double dC       = this->m_oParameters.GetC ();

	vec Y = Label;

	mat K = Data;

	colvec vDiags   = K.diag();
	int    N        = vDiags.n_elem;	

	ucolvec S;
	double  R = 0;

	colvec vAlpha;	
	colvec vDelta= (max(vDiags)-vDiags);

	mat mDiagC= diagmat(ones<vec>(N,1)/dC);
	mat mYY   = Y*Y.t();
	mat K2CVM = K % mYY +  mYY + mDiagC;
	vDiags    = K2CVM.diag();
	double iResult   = CoreSetSolver(K2CVM, dEpsilon,vDiags, vDelta, S, R, vAlpha);
	double dBias     = as_scalar(vAlpha.t()*Y(S));
	w.rows(S)   = vAlpha; 
	m_oCVMModel = cvm_model(w % Label,dBias,oKernel.GetKernelType()); 
}

cvm_parameters icvm_train::GetParameters(void)
{
	return m_oParameters;
}


int icvm_train::CoreSetSolver(const mat &K, double epsilon,const colvec &vDiags, colvec &vDeltas, ucolvec &S, double &R, colvec &vAlpha) 
{
	int    iResult = 0;
	int    N          = vDiags.n_elem;
	double dTolerance = (1+epsilon)*(1+epsilon);
	
	// default init
	ucolvec Sdefault = shuffle(linspace<ucolvec>(0,N-1,N));
	S                = Sdefault(uvec({0,1}));//randi<ucolvec>(2, distr_param(0,N-1) );    // choose 2 random points
	Sdefault         = Sdefault(find(Sdefault!=S(0))); // remove used entries
	Sdefault         = Sdefault(find(Sdefault!=S(1)));

//	std::cout << Sdefault.n_elem << std::endl;
	ucolvec SLast= S;
    vAlpha       = 0.5*ones<colvec>(2);                      // default loadings in case of 2 points (initial center)
	colvec vAlphaLast = vAlpha;
	mat mKSS     = K(S,S);

	R            = as_scalar((vAlpha.t() * (mKSS.diag()+vDeltas(S)) -vAlpha.t() * mKSS * vAlpha));   // thats R^2 - initial radius
	
	int iProbSampleSize          = 59;
	int NDefault                 = Sdefault.n_elem;
	ucolvec vDefaultRandSampling = shuffle(Sdefault);	
	ucolvec vRandSampling        = vDefaultRandSampling(linspace<ucolvec>(0,NDefault-1,min(iProbSampleSize,NDefault) ));//randi<ucolvec>(min(iProbSampleSize,N-1), distr_param(0,N-1) );
	
	mat mK       = K(S,vRandSampling);  // kernel of core points to all out of radius points
	mKSS         = K(S,S);  			// SFull is without S

	//vec vDistances = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + ones<vec>(vRandSampling.n_elem).t()*as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
	vec vDistances = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
	uvec vOutOfRadius = find(vDistances > dTolerance * R);

	// loop  until MEB is found
	int iCounter     = 2;
	double Rold         = R;
	iqpsolver	   iqpSolverInstance;
	while(vOutOfRadius.n_elem>0)
	{		
		uword I;
		vDistances(vOutOfRadius).max(I);                         	   // largest point away - original core-set heuristic
        uword iNewCoreIndex = vRandSampling(vOutOfRadius(I));    	   // identify new coreset point
		Sdefault            = Sdefault(find(Sdefault!=iNewCoreIndex)); // remove new added entry
//		std::cout << Sdefault.n_elem << "  " << iNewCoreIndex << std::endl;
		
		iCounter         = iCounter+1;
        SLast            = S;
		S				 = join_cols<umat>(S, uvec({iNewCoreIndex}));  // add to core-set		
        mKSS             = K(S,S);
        vAlphaLast       = vAlpha;
        int Ns           = S.n_elem;
        colvec f         = zeros<vec>(Ns); 
        vec ones1N       = ones<vec>(Ns);
        colvec zerosN1   = zeros<colvec>(Ns);
        colvec onesN1    = ones<colvec>(Ns);
        mKSS             = (mKSS + mKSS.t()) / 2.0;
		double dVal      = 0;

		mat fullA	     = diagmat(ones<vec>(Ns));				// alpha>=0
		fullA	         = join_cols(fullA, ones<vec>(Ns).t()); // sum alpha <= 1
		fullA	         = join_cols(fullA, ones<vec>(Ns).t()); // sum alpha >= 1
		colvec fullb     = zerosN1;								// alpha>=0		
	    fullb            = join_cols<mat>(fullb, vec({0}));		// sum alpha <= 1
		fullb            = join_cols<mat>(fullb, vec({-1}));	// sum alpha >= 1

		iResult          = iqpSolverInstance.InteriorPointSolver(2*mKSS, f, fullA, fullb, true, vAlpha, dVal);
		R                = as_scalar((vAlpha.t() * (mKSS.diag()+vDeltas(S)) -vAlpha.t() * mKSS * vAlpha));   // thats R^2 - initial radius
		vAlpha           = vAlpha/sum(vAlpha);

		NDefault                     = Sdefault.n_elem;
		ucolvec vDefaultRandSampling = shuffle(Sdefault);	
		vRandSampling                = vDefaultRandSampling(linspace<ucolvec>(0,NDefault-1,min(iProbSampleSize,NDefault) ));//randi<ucolvec>(min(iProbSampleSize,N-1), distr_param(0,N-1) );

		mK               = K(S,vRandSampling);
		vDistances       = (vDiags(vRandSampling).t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas(vRandSampling).t() ).t(); // get distances for MEB
		vOutOfRadius     = find(vDistances > dTolerance * R);
		//std::cout << sort(Sdefault.t()) << std::endl;
		//std::cout << "vR: " << vRandSampling.t() << std::endl;
		//std::cout << vOutOfRadius.n_elem << std::endl;
	}
	mK               = K.rows(S);
	vDistances       = (vDiags.t() -2 * vAlpha.t() * mK + as_scalar(vAlpha.t() * mKSS * vAlpha) + vDeltas.t() ).t(); // get distances for MEB
	vOutOfRadius     = find(vDistances > dTolerance * R);
//	std::cout << vDistances(vOutOfRadius) << " " << dTolerance * R << std::endl;
	return 0;
}

*/ // batch code
