/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * iqp_solver.cpp
 * Code is originally based on a Matlab code of G. Gordon see https://www.cs.cmu.edu/~ggordon/iqp.m
 * Copyright (C) 2015 Frank-Michael Schleif <schleify@cs.bham.ac.uk>
 *
 * This is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "iqp_solver.h"
#include <iostream>
#include <limits>
#include <armadillo>

#include <iostream>
#include <boost/bind.hpp>
#include <assert.h>
#include <fstream>
#include <boost/shared_ptr.hpp>


using namespace std;
using namespace arma;


int iqpsolver::InteriorPointSolver(const mat &H, const colvec &c, const mat &A, const vec &b, bool bVerbose, colvec &primals, double &dVal)
{
/*	std::cout << H << std::endl;
	std::cout << c << std::endl;
	std::cout << A << std::endl;
	std::cout << b << std::endl;*/
  // a few parameters
  double epsilon = 1e-8;    // accuracy requirement for solution
  double toobig = 1e+8;     // give up if solution gets too large
  double backoff = .99995;  // don't go quite all the way to boundary
  int maxiter = 75;      // stop after this many iterations
  double gamma = .5;        // how fast to try to decrease barrier
  double decay = .5;        //  -but effective gamma can be 1-(1-gamma)/decay
  double mingamma = 1e-8;   //  -but bounded below by mingamma
  int canreorder = 1;    // set to zero to prevent reordering

  int m = A.n_rows;
  int n = A.n_cols;
  

  // simple stupid initialization
  primals = zeros<colvec>(n,1);
  colvec duals = ones<colvec>(m,1);
  colvec slacks = ones<colvec>(m,1);
  double dComplementarity = as_scalar((slacks.t() * duals)) / m;

  // do until complementarity is small
  int iter = 0; double pstep = 1; double dstep = 1; int step = 0; double barrier = 100;

  while ((dComplementarity > epsilon) & (iter < maxiter))
  {
	    // target barrier is a fraction of complementarity determined
		// by how recently we've run against the edge of feasibility
		// (exact code here is arbitrary, anything reasonable should work)
		step = decay * min(pstep, dstep) * (step + 1/decay);
		double effgamma = max(mingamma, (1 - step * (1 - gamma)));
		barrier = min(barrier, dComplementarity * effgamma);

	    // compute rhs for Newton equations
		colvec invslacks = 1.0 / slacks;
		// d = spdiags(duals .* invslacks, 0, m, m);
	    mat d   = diagmat(duals % invslacks);
	    //std::cout << A << ":" << barrier << ":" << invslacks << ":" << duals << ":" << d << ":" << b << ":" << c << std::endl;
	    vec rhs = A.t() * (barrier * invslacks + duals - d * b) + c;

	    // solve Newton equations -- matlab is smart enough to use
		// Cholesky decomposition here
		//std::cout << (H + (A.t() * d * A)) << std::endl;
		//std::cout << rhs << std::endl;
		
		///colvec newprimals = solve((H + (A.t() * d * A)), rhs); //, solve_opts::fast);
		// -- that may fail due to the cholesky decomposition
		colvec newprimals;
		try
		{
			mat Ui = inv(  (chol(H + (A.t() * d * A)) ));
			newprimals = Ui*Ui.t() * rhs;
		}
		catch(...) 
		{
			newprimals = solve((H + (A.t() * d * A)), rhs); //, solve_opts::fast);
		}

		// from newprimals, compute newslacks
		colvec newslacks = A * newprimals + b;

		// from newprimals and newslacks, compute newduals
		colvec newduals = barrier * invslacks - d * (newslacks - slacks);

		// compute primal and dual stepsizes: longest possible w/o hitting zero
		pstep = 1 / max(max(slacks - newslacks, slacks) / slacks);
		dstep = 1 / max(max(duals - newduals, duals) / duals);

		if (newslacks.min() <= 0) pstep = backoff * pstep; 
		if (newduals.min() <= 0) dstep = backoff * dstep; 

		// do the update
		primals = primals + pstep * (newprimals - primals);
		slacks = slacks + pstep * (newslacks - slacks);
		duals = duals + dstep * (newduals - duals);
		dComplementarity = as_scalar((slacks.t() * duals)) / m;
		iter = iter + 1;

		// give up if numbers get too big
		if (max(max(slacks),max(duals)) > toobig)
		{
			dVal = std::numeric_limits<double>::quiet_NaN();
		
			if (bVerbose)
  				std::cout << "Warning: program appears to be ill-posed" << std::endl;
			return -1;
		}	  
  }

  dVal = as_scalar( primals.t() * c - primals.t() * H * primals / 2);
//	std::cout << dVal << " "
/*  std::cout << "*****************" << std::endl;
  std::cout	<< primals << std::endl;
  std::cout << "*****************" << std::endl;
  std::cout << "[ " << accu(primals) << " ] " << std::endl;
  */ return 0;
}

